"""Download necessary wheels and build the `coloraide` wheel."""
import sys
import subprocess
import os
import urllib.request
import glob
import shutil

urls = [
    "https://files.pythonhosted.org/packages/6e/33/1ae0f71395e618d6140fbbc9587cc3156591f748226075e0f7d6f9176522/Markdown-3.3.4-py3-none-any.whl",  # noqa: E501
    "https://files.pythonhosted.org/packages/9a/9a/36f71797fbaf1f4b6cb8debe1ab6d3ec969e8dbc651181f131a16b794d80/pymdown_extensions-9.0-py3-none-any.whl"  # noqa: E501
]
output = 'docs/src/markdown/playground/'
externals = {}
for url in urls:
    externals[os.path.join(output, url.split('/')[-1])] = url


def build_coloraide():
    """Build `coloraide` wheel."""
    cmd = [sys.executable, '-m', 'build', '--wheel', '-o', output]

    if sys.platform.startswith('win'):
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        process = subprocess.Popen(
            cmd,
            startupinfo=startupinfo,
            shell=False,
            env=os.environ.copy()
        )
    else:
        process = subprocess.Popen(
            cmd,
            shell=False,
            env=os.environ.copy()
        )
    process.communicate()

    return process.returncode


def download_wheel(url, dest):
    """Download a wheel."""

    print('Downloading: {}'.format(url))
    response = urllib.request.urlopen(url)
    if response.status != 200:
        print('Failed to download!')
        return 1
    with open(file, 'wb') as f:
        print('Writing: {}'.format(dest))
        f.write(response.read())


if __name__ == "__main__":
    # Clean up all old wheels
    for file in glob.glob(output + '*.whl'):
        if file not in externals.keys():
            os.remove(file)

    # Clean up build directory
    if os.path.exists('build'):
        shutil.rmtree('build')

    # Build `coloraide` wheel
    if build_coloraide():
        sys.exit(1)

    for file, url in externals.items():
        if download_wheel(url, file):
            sys.exit(1)

    sys.exit(0)



"""Porter Duff compositing."""
from abc import ABCMeta, abstractmethod
from typing import Type


class PorterDuff(metaclass=ABCMeta):
    """Porter Duff compositing."""

    def __init__(self, cba: float, csa: float) -> None:
        """Initialize."""

        self.cba = cba
        self.csa = csa

    @abstractmethod
    def _fa(self) -> float:  # pragma: no cover
        """Calculate `Fa`."""

        raise NotImplementedError('fa is not implemented')

    @abstractmethod
    def _fb(self) -> float:  # pragma: no cover
        """Calculate `Fb`."""

        raise NotImplementedError('fb is not implemented')

    def co(self, cb: float, cs: float) -> float:
        """Calculate premultiplied coordinate."""

        return self.csa * self._fa() * cs + self.cba * self._fb() * cb

    def ao(self) -> float:
        """Calculate output alpha."""

        return self.csa * self._fa() + self.cba * self._fb()


class Clear(PorterDuff):
    """Clear."""

    def _fa(self) -> float:
        """Calculate `Fa`."""

        return 0

    def _fb(self) -> float:
        """Calculate `Fb`."""

        return 0


class Copy(PorterDuff):
    """Copy."""

    def _fa(self) -> float:
        """Calculate `Fa`."""

        return 1

    def _fb(self) -> float:
        """Calculate `Fb`."""

        return 0


class Destination(PorterDuff):
    """Destination."""

    def _fa(self) -> float:
        """Calculate `Fa`."""

        return 0

    def _fb(self) -> float:
        """Calculate `Fb`."""

        return 1


class SourceOver(PorterDuff):
    """Source over."""

    def _fa(self) -> float:
        """Calculate `Fa`."""

        return 1

    def _fb(self) -> float:
        """Calculate `Fb`."""

        return 1 - self.csa


class DestinationOver(PorterDuff):
    """Destination over."""

    def _fa(self) -> float:
        """Calculate `Fa`."""

        return 1 - self.cba

    def _fb(self) -> float:
        """Calculate `Fb`."""

        return 1


class SourceIn(PorterDuff):
    """Source in."""

    def _fa(self) -> float:
        """Calculate `Fa`."""

        return self.cba

    def _fb(self) -> float:
        """Calculate `Fb`."""

        return 0


class DestinationeIn(PorterDuff):
    """Destination in."""

    def _fa(self) -> float:
        """Calculate `Fa`."""

        return 0

    def _fb(self) -> float:
        """Calculate `Fb`."""

        return self.csa


class SourceOut(PorterDuff):
    """Source out."""

    def _fa(self) -> float:
        """Calculate `Fa`."""

        return 1 - self.cba

    def _fb(self) -> float:
        """Calculate `Fb`."""

        return 0


class DestinationOut(PorterDuff):
    """Destination out."""

    def _fa(self) -> float:
        """Calculate `Fa`."""

        return 0

    def _fb(self) -> float:
        """Calculate `Fb`."""

        return 1 - self.csa


class SourceAtop(PorterDuff):
    """Source atop."""

    def _fa(self) -> float:
        """Calculate `Fa`."""

        return self.cba

    def _fb(self) -> float:
        """Calculate `Fb`."""

        return 1 - self.csa


class DestinationAtop(PorterDuff):
    """Destination atop."""

    def _fa(self) -> float:
        """Calculate `Fa`."""

        return 1 - self.cba

    def _fb(self) -> float:
        """Calculate `Fb`."""

        return self.csa


class XOR(PorterDuff):
    """XOR."""

    def _fa(self) -> float:
        """Calculate `Fa`."""

        return 1 - self.cba

    def _fb(self) -> float:
        """Calculate `Fb`."""

        return 1 - self.csa


class Lighter(PorterDuff):
    """Lighter."""

    def _fa(self) -> float:
        """Calculate `Fa`."""

        return 1

    def _fb(self) -> float:
        """Calculate `Fb`."""

        return 1


SUPPORTED = {
    'clear': Clear,
    'copy': Copy,
    'destination': Destination,
    'source-over': SourceOver,
    'destination-over': DestinationOver,
    'source-in': SourceIn,
    'destination-in': DestinationeIn,
    'source-out': SourceOut,
    'destination-out': DestinationOut,
    'source-atop': SourceAtop,
    'destination-atop': DestinationAtop,
    'xor': XOR,
    'lighter': Lighter
}


def compositor(name: str) -> Type[PorterDuff]:
    """Get the requested compositor."""

    name = name.lower()
    if name not in SUPPORTED:
        raise ValueError("'{}' compositing is not supported".format(name))
    return SUPPORTED[name]



"""
Calculate the Lch threshold of Lch-ish colors.

Particularly, we are interested in seeing the greatest
chroma deviation from zero for achromatic colors.

We simply will take a number of RGB color spaces and loop
through all achromatic colors, significantly past white
to see how far from zero chroma deviates.

This is mainly due to the fact that many of these Lch-ish
color spaces, as they get close to zero, whether due to
the algorithm, the precision, floating point math, or
whatever the reason, the conversion to an achromatic color
in the Lch-ish space often is very, very close to zero, but
not quite.

The threshold is essentially used during conversion to round
off chroma if it is very, very close to zero giving better,
expected conversions.

Looking at its Lab counterpart's `a` and `b` value is optional.
This is mainly if you are curious to see how the other channels
respond, or if you wanted use `a` and `b` to determine achromatic
response.
"""
import sys
import argparse
import os
import re

sys.path.insert(0, os.getcwd())

from coloraide import Color  # noqa: E402

RE_LEAD_ZERO = re.compile(r'^0\.0+')


def run(lch, lab):
    """Run the calculation."""

    for space in ('srgb', 'display-p3', 'rec2020', 'a98-rgb', 'prophoto-rgb'):
        max_chroma = 0.0
        max_a = 0.0
        max_b = 0.0

        for x in range(-0x600, 0x600):
            # Create an achromatic RGB color
            color = Color('color({space} {num:f} {num:f} {num:f})'.format(space=space, num=x / 255))
            if lab:
                labish = color.convert(lab)
                a_name, b_name = labish._space.labish_names()[1:]
                a = labish.get(a_name)
                b = labish.get(b_name)
                if abs(a) > max_a:
                    max_a = abs(a)
                if abs(b) > max_b:
                    max_b = abs(b)

            if lch:
                lchish = color.convert(lch)
                c_name = lchish._space.lchish_names()[1]
                chroma = lchish.get(c_name)
                if chroma > max_chroma:
                    max_chroma = chroma

    if lab:
        print('{}: maximum a: {:.53f}'.format(lab, max_a))
        print('{}: maximum b: {:.53f}'.format(lab, max_b))

    if lch:
        num = '{:.53f}'.format(max_chroma)
        m = RE_LEAD_ZERO.match(num)
        minimum = '0'
        better = '0'
        if len(m.group(0)) != len(num):
            count = len(m.group(0)[2:])
            if int(num[m.end(0)]) == 9:
                # Close to rolling over
                minimum = '> 0.{}1'.format('0' * (count - 1))
            else:
                # Less than a decimal point of room
                minimum = '> 0.{}{}'.format('0' * count, str(int(num[m.end(0)]) + 1))
            # Give at least a little over a decimal point
            better = '> 0.{}2'.format('0' * (count - 1))
        print('{}: maximum chroma: {:.53f}'.format(lch, max_chroma))
        print('{}: minimum threshold: {}'.format(lch, minimum))
        print('{}: relaxed threshold: {}'.format(lch, better))

        print(
            '\n* Only potential recommendations, adjustments can be made if desired.'
            '\n  If consraints are very tight, the minimum is probably best while the'
            '\n  relaxed would give a little more room assuming achromatic colors even'
            '\n  further out rolled the next decimal point by one.'
        )

    return 0


def main():
    """Main."""

    parser = argparse.ArgumentParser(
        prog='calc_lch_threshold.py', description='Calculate achromatic threshold for Lch-ish colors.'
    )
    # Flag arguments
    parser.add_argument(
        '--lch', '-c', action='store', default='',
        help="The Lch color whose 'chroma' values you'd like to evaluate."
    )
    parser.add_argument(
        '--lab', '-l', action='store', default='',
        help="Optionally view Lab color whose 'ab' values you'd like evaluate."
    )
    args = parser.parse_args()

    if not args.lch and not args.lab:
        print('No spaces provided to test!')

    return run(args.lch, args.lab)


if __name__ == "__main__":
    sys.exit(main())



"""Test Jzazbz library."""
import unittest
from . import util
from coloraide import Color


class TestJzazbzInputOutput(util.ColorAsserts, unittest.TestCase):
    """Test Jzazbz."""

    def test_input_raw(self):
        """Test raw input."""

        self.assertColorEqual(Color("jzazbz", [1, 0.5, 0.5]), Color('color(--jzazbz 100% 0.5 0.5)'))

    def test_color_class(self):
        """Test raw input."""

        self.assertColorEqual(Color(Color("jzazbz", [1, 0.5, 0.5])), Color('color(--jzazbz 100% 0.5 0.5)'))

    def test_color(self):
        """Test color input/output format."""

        args = {"color": True}
        color = "color(--jzazbz 1 0.2 -0.3)"

        self.assertEqual(Color(color).to_string(**args), 'color(--jzazbz 1 0.2 -0.3)')

        color = "color(--jzazbz 1 0.2 -0.3 / 0.5)"
        self.assertEqual(Color(color).to_string(**args), 'color(--jzazbz 1 0.2 -0.3 / 0.5)')

        color = "color(--jzazbz 100% 0.2 -0.3 / 50%)"
        self.assertEqual(Color(color).to_string(**args), 'color(--jzazbz 1 0.2 -0.3 / 0.5)')

    def test_no_alpha(self):
        """Test no alpha."""

        args = {"alpha": False}

        color = "color(--jzazbz 1 0.2 -0.3 / 0.5)"
        jzazbz = Color(color)
        self.assertEqual("color(--jzazbz 1 0.2 -0.3)", jzazbz.to_string(**args))

    def test_force_alpha(self):
        """Test force alpha."""

        args = {"alpha": True}

        color = "color(--jzazbz 1 0.2 -0.3 / 100%)"
        jzazbz = Color(color)
        self.assertEqual("color(--jzazbz 1 0.2 -0.3 / 1)", jzazbz.to_string(**args))

    def test_precision(self):
        """Test precision."""

        color = 'color(--jzazbz 0.123456 0.123456 -0.123456)'
        self.assertEqual(Color(color).to_string(), 'color(--jzazbz 0.12346 0.12346 -0.12346)')
        self.assertEqual(Color(color).to_string(precision=3), 'color(--jzazbz 0.123 0.123 -0.123)')
        self.assertEqual(Color(color).to_string(precision=0), 'color(--jzazbz 0 0 0)')
        self.assertEqual(
            Color(color).to_string(precision=-1),
            'color(--jzazbz 0.12345599999999999629718416827017790637910366058349609 0.12345599999999999629718416827017790637910366058349609 -0.12345599999999999629718416827017790637910366058349609)'  # noqa:  E501
        )

    def test_fit(self):
        """Test fit."""

        self.assertEqual(
            Color('color(--jzazbz 2 0.6 -0.6)').to_string(),
            'color(--jzazbz 2 0.6 -0.6)'
        )

        self.assertEqual(
            Color('color(--jzazbz 2 0.6 -0.6)').to_string(fit="clip"),
            'color(--jzazbz 2 0.6 -0.6)'
        )

        self.assertEqual(
            Color('color(--jzazbz 2 0.6 -0.6)').to_string(fit=False),
            'color(--jzazbz 2 0.6 -0.6)'
        )


class TestJzazbzProperties(util.ColorAsserts, unittest.TestCase):
    """Test Jzazbz."""

    def test_jz(self):
        """Test `lightness`."""

        c = Color('color(--jzazbz 1 0.2 -0.3 / 1)')
        self.assertEqual(c.jz, 1)
        c.jz = 0.2
        self.assertEqual(c.jz, 0.2)

    def test_az(self):
        """Test `az`."""

        c = Color('color(--jzazbz 1 0.2 -0.3 / 1)')
        self.assertEqual(c.az, 0.2)
        c.az = 0.1
        self.assertEqual(c.az, 0.1)

    def test_bz(self):
        """Test `bz`."""

        c = Color('color(--jzazbz 1 0.2 -0.3 / 1)')
        self.assertEqual(c.bz, -0.3)
        c.bz = 0.1
        self.assertEqual(c.bz, 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--jzazbz 1 0.2 -0.3 / 1)')
        self.assertEqual(c.alpha, 1)
        c.alpha = 0.5
        self.assertEqual(c.alpha, 0.5)



"""
Render 3D plots of sRGB in other color spaces.

In order to render things fast and allow for reasonable
performance, we render the outer shell of the space.

Only sRGB and related sRGB cylindrical models are precise,
all others are approximations of the shape due to our
approach.
"""
import itertools
import matplotlib.pyplot as plt
import sys
import os
import math
import argparse

sys.path.insert(0, os.getcwd())

from coloraide import Color  # noqa: E402
from coloraide.spaces import Cylindrical, Lchish, Labish  # noqa: E402
from coloraide.util import is_nan  # noqa: E402

axis_map = {
    # Lab like spaces
    "lab": [1, 2, 0],
    "lab-d65": [1, 2, 0],
    "oklab": [1, 2, 0],
    "jzazbz": [1, 2, 0],
    "ictcp": [1, 2, 0],
    "din99o": [1, 2, 0],
    "luv": [1, 2, 0],
    "luv-d65": [1, 2, 0],

    # Lch like spaces
    "lch": [2, 1, 0],
    "lch-d65": [2, 1, 0],
    "oklch": [2, 1, 0],
    "jzczhz": [2, 1, 0],
    "din99o-lch": [2, 1, 0],
    "lchuv": [2, 1, 0],
    "lchuv-d65": [2, 1, 0]
}


def add_color(space, color, x, y, z, c):
    """Add color to the provided arrays."""

    coords = color.convert(space).coords()
    x.append(coords[0])
    y.append(coords[1])
    z.append(coords[2])
    s = color.to_string(hex=True)
    c.append(s)


def add_cyl_color(space, color, x, y, z, c):
    """
    Add color to the provided arrays.

    Handles cylindrical spaces. Returns x (hue), y (chroma/saturation), z (value/lightness).
    """

    cyl = color.convert(space)

    chroma = cyl.chroma
    lightness = cyl.lightness
    hue = cyl.hue

    if is_nan(hue):
        hue = 0

    z.append(chroma * math.sin(math.radians(hue)))
    y.append(chroma * math.cos(math.radians(hue)))
    x.append(lightness)

    s = color.convert('srgb').to_string(hex=True)
    c.append(s)


def render_cart_space(space, resolution, data, c):
    """Render the space with the given resolution and factor."""

    x, y, z = data

    # We are rendering the spaces using sRGB, so just do a shell by picking
    # all the colors on the outside of the sRGB space. Render will be hollow.
    color = Color('srgb', [])
    for c1, c2 in itertools.product(
        (x / resolution for x in range(0, resolution + 1)),
        (x / resolution for x in range(0, resolution + 1))
    ):

        add_color(space, color.update('srgb', [0, c1, c2]), x, y, z, c)
        add_color(space, color.update('srgb', [1, c1, c2]), x, y, z, c)

        add_color(space, color.update('srgb', [c1, 0, c2]), x, y, z, c)
        add_color(space, color.update('srgb', [c1, 1, c2]), x, y, z, c)

        add_color(space, color.update('srgb', [c1, c2, 0]), x, y, z, c)
        add_color(space, color.update('srgb', [c1, c2, 1]), x, y, z, c)


def render_cyl_space(space, resolution, data, c):
    """
    Render the space with the given resolution and factor.

    The sRGB gamut is used for the plot. The HSV space is used as it maps
    far better to cylinder spaces. Very close to black, on some spaces the
    models don't cover very well, so we make an additional pass using
    Cartesian coordinates extremely close to black.

    It can be noted, we don't bother plotting anything from the bottom
    disc of the HSV cylinder as they all resolve to pure black. It generates
    a lot of redundant points.
    """

    x, y, z = data

    # Resolution increase in non-hue channels helps smooth out some spaces a bit more.
    res2 = int(resolution * 1.5)

    color = Color('srgb', [])
    is_labish = issubclass(Color.CS_MAP[space], Labish)
    add = add_color if is_labish else add_cyl_color

    # We are rendering the spaces using sRGB, so just do a shell by picking
    # all the colors on the outside of the sRGB space. Render will be hollow.
    for c1, c2 in itertools.product(
        ((x / resolution) * 360 for x in range(0, resolution + 1)),
        ((x / res2) for x in range(0, res2 + 1))
    ):

        # Only the top disc provides useful points, everything in the bottom just yields black.
        add(space, color.update('hsv', [c1, c2, 1]), x, y, z, c)
        add(space, color.update('hsv', [c1, 1, c2]), x, y, z, c)
        add(space, color.update('hsv', [c1, 1, c2 * 0.005]), x, y, z, c)


def render_srgb_cyl_space(space, resolution, data, c):
    """
    Render the sRGB cylindrical space: HSL, HSV, HWB, etc.

    Will render cylinder with the caps on the top and bottom.
    """

    x, y, z = data

    # Render the cylinder by iterating through the hues and mapping them at the farthest
    # point from the center creating a hollow cylinder. Also, render the top and bottom disc caps.
    color = Color("srgb", [])
    for c1, t in itertools.product(
        ((x / resolution) * 360 for x in range(0, resolution + 1)),
        (((x / resolution), i) for i, x in enumerate(range(0, resolution + 1), 0))
    ):

        # Offset the plot on every other iteration blend the rows into a mesh
        # Better looking when low resolution zoomed into higher resolution
        c2, count = t
        if count % 2 and c1 < 360:
            c1 += (360 / resolution) * 0.5

        # Top disc
        x.append(c2 * math.sin(math.radians(c1)))
        y.append(c2 * math.cos(math.radians(c1)))
        z.append(1)
        c.append(color.update(space, [c1, c2, 1]).to_string(hex=True))

        # Bottom disc
        x.append(c2 * math.sin(math.radians(c1)))
        y.append(c2 * math.cos(math.radians(c1)))
        z.append(0)
        c.append(color.update(space, [c1, c2, 0]).to_string(hex=True))

        # Cylinder portion
        x.append(1 * math.sin(math.radians(c1)))
        y.append(1 * math.cos(math.radians(c1)))
        z.append(c2)
        c.append(color.update(space, [c1, 1, c2]).to_string(hex=True))


def plot_space_in_srgb(space, title="", dark=False, resolution=70):
    """Plot the given space in sRGB."""

    data = [[], [], []]
    c = []

    # Get names for
    names = Color.CS_MAP[space].CHANNEL_NAMES
    is_cyl = issubclass(Color.CS_MAP[space], Cylindrical)
    is_labish = issubclass(Color.CS_MAP[space], Labish)
    is_srgb_cyl = is_cyl and not issubclass(Color.CS_MAP[space], Lchish)

    # Some spaces need us to rearrange the order of the data
    axm = axis_map.get(space, [0, 1, 2])

    # Select the right theme
    if dark:
        plt.style.use('dark_background')
    else:
        plt.style.use('seaborn-whitegrid')

    # Setup figure and axis
    figure = plt.figure()
    plt.tight_layout()
    ax = plt.axes(
        projection='3d',
        xlabel=names[axm[0]] if not is_cyl else "{} (0˚ - 360˚)".format(names[axm[0]]),
        ylabel=names[axm[1]],
        zlabel=names[axm[2]]
    )
    # Turn off ticks for cylindrical hue
    if is_cyl:
        ax.xaxis.set_ticks([])
    figure.add_axes(ax)

    # Add title
    plt.title(title if title else 'srgb rendered in {}'.format(space), pad=20)

    # Render the space
    if is_srgb_cyl:
        # Render a sRGB cylinder style plot
        render_srgb_cyl_space(space, resolution, data, c)
    elif is_labish or is_cyl:
        # Render cylindrical spaces. Lab like spaces are cylindrical,
        # just represented in the Cartesian coordinate system.
        render_cyl_space(space, resolution, data, c)
    else:
        # Render Cartesian spaces.
        # These are rectangular spaces like RGB spaces and XYZ.
        render_cart_space(space, resolution, data, c)

    # Setup the aspect ratio
    ax.set_box_aspect((1, 1, 1))

    # Plot the data
    ax.scatter3D(data[axm[0]], data[axm[1]], data[axm[2]], c=c, s=20 * 4)


def main():
    """Main."""

    parser = argparse.ArgumentParser(prog='3d_diagrams', description='Plot 3D sRGB in different color spaces.')
    parser.add_argument('--space', '-s', help='Desired space.')
    parser.add_argument(
        '--resolution', '-r',
        default="200",
        help=(
            "How densely to render the figure. Some spaces need higher resolution to flesh out certain areas, "
            "but it comes at the cost of speed."
        )
    )
    parser.add_argument('--title', '-t', default='', help="Provide a title for the diagram.")
    parser.add_argument('--dark', action="store_true", help="Use dark theme.")
    parser.add_argument('--output', '-o', default='', help='Output file.')
    args = parser.parse_args()

    plot_space_in_srgb(
        args.space,
        title=args.title,
        dark=args.dark,
        resolution=int(args.resolution)
    )

    if args.output:
        plt.savefig(args.output)
    else:
        plt.show()


if __name__ == "__main__":
    sys.exit(main())



"""LCH class."""
from ..spaces import RE_DEFAULT_MATCH
from .lchuv import Lchuv
import re


class LchuvD65(Lchuv):
    """Lch(uv) class."""

    BASE = "luv-d65"
    NAME = "lchuv-d65"
    SERIALIZE = ("--lchuv-d65",)
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"



"""Test Okhsl library."""
from coloraide.util import NaN
import unittest
from . import util
from coloraide import Color


class TestOkhslInputOutput(util.ColorAsserts, unittest.TestCase):
    """Test Okhsl."""

    def test_input_raw(self):
        """Test raw input."""

        self.assertColorEqual(Color("okhsl", [20, 1, 0.75]), Color('color(--okhsl 20 1 0.75)'))

    def test_conversion(self):
        """
        Test a particular area of the conversion algorithm to ensure complete coverage.

        Exercises part where chroma is less than calculated mid chroma.
        """

        self.assertEqual(
            Color('color(--oklab 0.5 0.1 0)').convert('okhsl').to_string(),
            'color(--okhsl 0 0.49656 0.42114)'
        )

    def test_color_class(self):
        """Test raw input."""

        self.assertColorEqual(Color(Color("okhsl", [20, 1, 0.75])), Color('color(--okhsl 20 1 0.75)'))

    def test_percent(self):
        """Test that percents work properly."""

        color = "color(--okhsl 20 100% 75% / 100%)"
        okhsl = Color(color)
        self.assertEqual("color(--okhsl 20 1 0.75)", okhsl.to_string())

        color = "color(--okhsl 20 1 0.75 / 20%)"
        okhsl = Color(color)
        self.assertEqual("color(--okhsl 20 1 0.75 / 0.2)", okhsl.to_string())

    def test_no_alpha(self):
        """Test no alpha."""

        args = {"alpha": False}

        color = "color(--okhsl 20 100% 75% / 0.2)"
        okhsl = Color(color)
        self.assertEqual("color(--okhsl 20 1 0.75)", okhsl.to_string(**args))

    def test_force_alpha(self):
        """Test force alpha."""

        args = {"alpha": True}

        color = "color(--okhsl 20 100% 75% / 1)"
        okhsl = Color(color)
        self.assertEqual("color(--okhsl 20 1 0.75 / 1)", okhsl.to_string(**args))

    def test_precision(self):
        """Test precision."""

        color = 'color(--okhsl 20.1234567 0.1234567 0.1234567)'
        self.assertEqual(Color(color).to_string(), 'color(--okhsl 20.123 0.12346 0.12346)')
        self.assertEqual(Color(color).to_string(precision=3), 'color(--okhsl 20.1 0.123 0.123)')
        self.assertEqual(Color(color).to_string(precision=0), 'color(--okhsl 20 0 0)')
        self.assertEqual(
            Color(color).to_string(precision=-1),
            'color(--okhsl 20.12345669999999842048055143095552921295166015625 0.12345670000000000254836152180359931662678718566894531 0.12345670000000000254836152180359931662678718566894531)'  # noqa:  E501
        )

    def test_fit(self):
        """Test fit."""

        self.assertEqual(
            Color('color(--okhsl 50 40% 110%)').to_string(),
            'color(--okhsl 0 0 1)'
        )

        self.assertEqual(
            Color('color(--okhsl 50 40% 110%)').to_string(fit="clip"),
            'color(--okhsl 50 0.4 1)'
        )

        self.assertEqual(
            Color('color(--okhsl 50 40% 110%)').to_string(fit=False),
            'color(--okhsl 50 0.4 1.1)'
        )


class TestOkhslProperties(util.ColorAsserts, unittest.TestCase):
    """Test Okhsl."""

    def test_hue(self):
        """Test `hue`."""

        c = Color('color(--okhsl 120 50% 90% / 1)')
        self.assertEqual(c.hue, 120)
        c.hue = 110
        self.assertEqual(c.hue, 110)

    def test_saturation(self):
        """Test `saturation`."""

        c = Color('color(--okhsl 120 50% 90% / 1)')
        self.assertEqual(c.saturation, 0.5)
        c.saturation = 0.6
        self.assertEqual(c.saturation, 0.6)

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--okhsl 120 50% 90% / 1)')
        self.assertEqual(c.lightness, 0.9)
        c.lightness = 0.8
        self.assertEqual(c.lightness, 0.8)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--okhsl 120 50% 90% / 1)')
        self.assertEqual(c.alpha, 1)
        c.alpha = 0.5
        self.assertEqual(c.alpha, 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('okhsl', [NaN, 0.5, 1], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_auto_null(self):
        """Test auto null."""

        c = Color('color(--okhsl 120 0% 75% / 1)')
        self.assertTrue(c.is_nan('hue'))



"""Delta E 76."""
from ..distance import DeltaE, distance_euclidean
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class DE76(DeltaE):
    """Delta E 76 class."""

    NAME = "76"
    SPACE = "lab"

    @classmethod
    def distance(cls, color: 'Color', sample: 'Color', **kwargs: Any) -> float:
        """
        Delta E 1976 color distance formula.

        http://www.brucelindbloom.com/index.html?Eqn_DeltaE_CIE76.html

        Basically this is Euclidean distance in the Lab space.
        """

        # Equation (1)
        return distance_euclidean(color, sample, space=cls.SPACE)



"""XYZ D65 class."""
from ..spaces import Space, RE_DEFAULT_MATCH, GamutUnbound
import re
from ..util import MutableVector
from typing import Tuple


class XYZD65(Space):
    """XYZ D65 class."""

    BASE = "xyz-d65"
    NAME = "xyz-d65"
    SERIALIZE = ("xyz-d65", 'xyz')  # type: Tuple[str, ...]
    CHANNEL_NAMES = ("x", "y", "z")
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"

    BOUNDS = (
        GamutUnbound(0.0, 1.0),
        GamutUnbound(0.0, 1.0),
        GamutUnbound(0.0, 1.0)
    )

    @property
    def x(self) -> float:
        """X channel."""

        return self._coords[0]

    @x.setter
    def x(self, value: float) -> None:
        """Shift the X."""

        self._coords[0] = self._handle_input(value)

    @property
    def y(self) -> float:
        """Y channel."""

        return self._coords[1]

    @y.setter
    def y(self, value: float) -> None:
        """Set Y."""

        self._coords[1] = self._handle_input(value)

    @property
    def z(self) -> float:
        """Z channel."""

        return self._coords[2]

    @z.setter
    def z(self, value: float) -> None:
        """Set Z channel."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """
        To XYZ (no change).

        Any needed chromatic adaptation is handled in the parent Color object.
        """

        return coords

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """
        From XYZ (no change).

        Any needed chromatic adaptation is handled in the parent Color object.
        """

        return coords



"""Delta E 94."""
from ..distance import DeltaE
import math
from .. import util
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class DE94(DeltaE):
    """Delta E 94 class."""

    NAME = "94"

    @classmethod
    def distance(
        cls,
        color: 'Color',
        sample: 'Color',
        kl: float = 1,
        k1: float = 0.045,
        k2: float = 0.015,
        **kwargs: Any
    ) -> float:
        """
        Delta E 1994 color distance formula.

        http://www.brucelindbloom.com/Eqn_DeltaE_CIE94.html
        """

        l1, a1, b1 = util.no_nans(color.convert("lab").coords())
        l2, a2, b2 = util.no_nans(sample.convert("lab").coords())

        # Equation (5)
        c1 = math.sqrt(a1 ** 2 + b1 ** 2)

        # Equation (6)
        c2 = math.sqrt(a2 ** 2 + b2 ** 2)

        # Equation  (2)
        dl = l1 - l2

        # Equation  (3)
        dc = c1 - c2

        # Equation (7)
        da = a1 - a2

        # Equation  (8)
        db = b1 - b2

        # Equation (4)
        # We never reference `dh` until the very end, and when we do, we square it
        # before using it, so we don't need the square root as described in the
        # algorithm. Instead we can just leave the result as is.
        dh = da ** 2 + db ** 2 - dc ** 2

        # Equation (9)
        sl = 1

        # Equation (10)
        sc = 1 + k1 * c1

        # Equation (11)
        sh = 1 + k2 * c1

        # Equation (12)
        # Provided by `kl`

        # Equation (13)
        kc = 1

        # Equation (14)
        kh = 1

        # Equation (15) and Equation (16)
        # Provided by `k1` and `k2`

        # Equation (1)
        return math.sqrt(
            (dl / (kl * sl)) ** 2 +
            (dc / (kc * sc)) ** 2 +
            # Square root just the denominator as `dh` is already squared.
            dh / ((kh * sh) ** 2)
        )



"""LCH class."""
from ..spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, FLG_ANGLE, FLG_PERCENT
from .lch import Lch, ACHROMATIC_THRESHOLD
from .. import util
import re
import math
from ..util import MutableVector


def luv_to_lchuv(luv: MutableVector) -> MutableVector:
    """Luv to Lch(uv)."""

    l, u, v = luv

    c = math.sqrt(u ** 2 + v ** 2)
    h = math.degrees(math.atan2(v, u))

    # Achromatic colors will often get extremely close, but not quite hit zero.
    # Essentially, we want to discard noise through rounding and such.
    if c < ACHROMATIC_THRESHOLD:
        h = util.NaN

    return [l, c, util.constrain_hue(h)]


def lchuv_to_luv(lchuv: MutableVector) -> MutableVector:
    """Lch(uv) to Luv."""

    l, c, h = lchuv
    h = util.no_nan(h)

    # If, for whatever reason (mainly direct user input),
    # if chroma is less than zero, clamp to zero.
    if c < 0.0:
        c = 0.0

    return [
        l,
        c * math.cos(math.radians(h)),
        c * math.sin(math.radians(h))
    ]


class Lchuv(Lch, Space):
    """Lch(uv) class."""

    BASE = "luv"
    NAME = "lchuv"
    SERIALIZE = ("--lchuv",)
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D50"

    BOUNDS = (
        GamutUnbound(0, 100.0, FLG_PERCENT),
        GamutUnbound(0.0, 176.0),
        GamutUnbound(0.0, 360.0, FLG_ANGLE)
    )

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To Luv from Lch(uv)."""

        return lchuv_to_luv(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From Luv to Lch(uv)."""

        return luv_to_lchuv(coords)



"""
Din99o class.

https://de.wikipedia.org/wiki/DIN99-Farbraum
"""
from ..spaces import RE_DEFAULT_MATCH
from .lab import Lab
import re
import math
from ..util import MutableVector

KE = 1
KCH = 1

# --- DIN99o ---
# C1 was a bit off due to rounding and gave us less than 100
# lightness when translating from sRGB white.
# Solving the equation for a lightness of 100 and `KE` of 1,
# which should give us a 100 lightness for white:
#
# ```
# L99o = (303.67 * ln(1 + 0.0039 * L*)) / KE
# 303.67 = 100 / ln(1 + 0.0039 * 100)
# 303.67 = 100 / ln(1.39)
# ```
#
# This gives 303.67100547050995 without rounding and fixes
# white translation.
RADS = math.radians(26)
FACTOR = 0.83
C1 = 100 / math.log(1.39)
C2 = 0.0039
C3 = 0.075
C4 = 0.0435


def lab_to_din99o(lab: MutableVector) -> MutableVector:
    """XYZ to Din99o."""

    l, a, b = lab
    val = 1 + C2 * l
    l99o = C1 * math.copysign(math.log(abs(val)), val) / KE

    if a == 0 and b == 0:
        a99o = b99o = 0.0
    else:
        eo = a * math.cos(RADS) + b * math.sin(RADS)
        fo = FACTOR * (b * math.cos(RADS) - a * math.sin(RADS))
        go = math.sqrt(eo ** 2 + fo ** 2)
        val = 1 + C3 * go
        c99o = math.copysign(math.log(abs(val)), val) / (C4 * KE * KCH)
        h99o = math.atan2(fo, eo) + RADS

        a99o = c99o * math.cos(h99o)
        b99o = c99o * math.sin(h99o)

    return [l99o, a99o, b99o]


def din99o_lab_to_lch(lab: MutableVector) -> MutableVector:
    """
    Convert Din99o Lab to Lch.

    Hue is in radians.
    """

    l99o, a99o, b99o = lab
    h99o = math.atan2(b99o, a99o)
    c99o = math.sqrt(a99o ** 2 + b99o ** 2)

    return [l99o, c99o, h99o]


def din99o_to_lab(din99o: MutableVector) -> MutableVector:
    """Din99o to XYZ."""

    l99o, c99o, h99o = din99o_lab_to_lch(din99o)
    val = C4 * c99o * KCH * KE
    g = (math.copysign(math.exp(abs(val)), val) - 1) / C3
    e = g * math.cos(h99o - RADS)
    f = g * math.sin(h99o - RADS)

    val = (l99o * KE) / C1
    (math.copysign(math.exp(abs(val)), val) - 1) / C2

    return [
        (math.exp((l99o * KE) / C1) - 1) / C2,
        e * math.cos(RADS) - (f / FACTOR) * math.sin(RADS),
        e * math.sin(RADS) + (f / FACTOR) * math.cos(RADS)
    ]


class Din99o(Lab):
    """Din99o class."""

    BASE = 'xyz-d65'
    NAME = "din99o"
    SERIALIZE = ("--din99o",)
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To XYZ from Din99o."""

        return super().to_base(din99o_to_lab(coords))

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From XYZ to Din99o."""

        return lab_to_din99o(super().from_base(coords))



"""Parse utilities."""
import re
import math
from . import util

RGB_CHANNEL_SCALE = 1.0 / 255.0
HUE_SCALE = 1.0 / 360.0
SCALE_PERCENT = 1 / 100.0

CONVERT_TURN = 360
CONVERT_GRAD = 90 / 100

RE_CHAN_SPLIT = re.compile(r'(?:\s*[,/]\s*|\s+)')
RE_COMMA_SPlIT = re.compile(r'(?:\s*,\s*)')
RE_SLASH_SPLIT = re.compile(r'(?:\s*/\s*)')

COLOR_PARTS = {
    "strict_percent": r"(?:[+\-]?(?:(?:[0-9]*\.[0-9]+)|[0-9]+)(?:e[-+]?[0-9]*)?%)",
    "strict_float": r"(?:[+\-]?(?:(?:[0-9]*\.[0-9]+)|[0-9]+)(?:e[-+]?[0-9]*)?)",
    "percent": r"(?:[+\-]?(?:(?:[0-9]*\.[0-9]+)|[0-9]+)(?:e[-+]?[0-9]*)?%|none)",
    "float": r"(?:[+\-]?(?:(?:[0-9]*\.[0-9]+)|[0-9]+)(?:e[-+]?[0-9]*)?|none)",
    "angle": r"(?:[+\-]?(?:(?:[0-9]*\.[0-9]+)|[0-9]+)(?:e[-+]?[0-9]*)?(?:deg|rad|turn|grad)?|none)",
    "space": r"\s+",
    "comma": r"\s*,\s*",
    "slash": r"\s*/\s*",
    "sep": r"(?:\s*,\s*|\s+)",
    "asep": r"(?:\s*[,/]\s*|\s+)",
    "hex": r"[a-f0-9]"
}


def norm_float(string: str) -> float:
    """Normalize a float value."""

    if string == "none":
        return util.NaN
    elif string.lower().endswith(('e-', 'e+', 'e')):
        string += '0'
    return float(string)


def norm_hex_channel(string: str) -> float:
    """Normalize the hex string to a form we can handle."""

    if string.startswith('#'):
        return int(string[1:], 16) * RGB_CHANNEL_SCALE
    else:  # pragma: no cover
        # Should only occur internally if we are doing something wrong.
        raise ValueError("Unexpected value '{}'".format(string))


def norm_percent_channel(string: str, scale: bool = False) -> float:
    """Normalize percent channel."""

    if string == 'none':
        return norm_float(string)
    elif string.endswith('%'):
        value = norm_float(string[:-1])
        return value / 100.0 if scale else value
    else:  # pragma: no cover
        # Should only occur internally if we are doing something wrong.
        raise ValueError("Unexpected value '{}'".format(string))


def norm_color_channel(string: str, scale: bool = True) -> float:
    """Normalize percent channel."""

    if string.endswith('%'):
        value = norm_float(string[:-1])
        return value / 100.0 if scale else value
    else:
        return norm_float(string)


def norm_rgb_channel(string: str) -> float:
    """Normalize RGB channel."""

    if string.endswith("%"):
        return norm_percent_channel(string, True)
    else:
        return norm_float(string) * RGB_CHANNEL_SCALE


def norm_alpha_channel(string: str) -> float:
    """Normalize alpha channel."""

    if string.endswith("%"):
        value = norm_percent_channel(string, True)
    else:
        value = norm_float(string)
    return util.clamp(value, 0.0, 1.0)


def norm_angle_channel(angle: str) -> float:
    """Normalize angle units."""

    if angle.endswith('turn'):
        value = norm_float(angle[:-4]) * CONVERT_TURN
    elif angle.endswith('grad'):
        value = norm_float(angle[:-4]) * CONVERT_GRAD
    elif angle.endswith('rad'):
        value = math.degrees(norm_float(angle[:-3]))
    elif angle.endswith('deg'):
        value = norm_float(angle[:-3])
    else:
        value = norm_float(angle)
    return value



"""
JzCzhz class.

https://www.osapublishing.org/oe/fulltext.cfm?uri=oe-25-13-15131&id=368272
"""
from ..spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, Lchish, FLG_ANGLE, FLG_OPT_PERCENT
from .. import util
import re
import math
from ..util import MutableVector
from typing import Tuple

ACHROMATIC_THRESHOLD = 0.0003


def jzazbz_to_jzczhz(jzazbz: MutableVector) -> MutableVector:
    """Jzazbz to JzCzhz."""

    jz, az, bz = jzazbz

    cz = math.sqrt(az ** 2 + bz ** 2)
    hz = math.degrees(math.atan2(bz, az))

    # Achromatic colors will often get extremely close, but not quite hit zero.
    # Essentially, we want to discard noise through rounding and such.
    if cz < ACHROMATIC_THRESHOLD:
        hz = util.NaN

    return [jz, cz, util.constrain_hue(hz)]


def jzczhz_to_jzazbz(jzczhz: MutableVector) -> MutableVector:
    """JzCzhz to Jzazbz."""

    jz, cz, hz = jzczhz
    hz = util.no_nan(hz)

    # If, for whatever reason (mainly direct user input),
    # if chroma is less than zero, clamp to zero.
    if cz < 0.0:
        cz = 0.0

    return [
        jz,
        cz * math.cos(math.radians(hz)),
        cz * math.sin(math.radians(hz))
    ]


class JzCzhz(Lchish, Space):
    """
    JzCzhz class.

    https://www.osapublishing.org/oe/fulltext.cfm?uri=oe-25-13-15131&id=368272
    """

    BASE = "jzazbz"
    NAME = "jzczhz"
    SERIALIZE = ("--jzczhz",)
    CHANNEL_NAMES = ("jz", "cz", "hz")
    CHANNEL_ALIASES = {
        "lightness": "jz",
        "chroma": "cz",
        "hue": "hz"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"

    BOUNDS = (
        GamutUnbound(0.0, 1.0, FLG_OPT_PERCENT),
        GamutUnbound(0.0, 1.0),
        GamutUnbound(0.0, 360.0, FLG_ANGLE)
    )

    @property
    def jz(self) -> float:
        """Jz."""

        return self._coords[0]

    @jz.setter
    def jz(self, value: float) -> None:
        """Set jz."""

        self._coords[0] = self._handle_input(value)

    @property
    def cz(self) -> float:
        """Chroma."""

        return self._coords[1]

    @cz.setter
    def cz(self, value: float) -> None:
        """Set chroma."""

        self._coords[1] = self._handle_input(value)

    @property
    def hz(self) -> float:
        """Hue."""

        return self._coords[2]

    @hz.setter
    def hz(self, value: float) -> None:
        """Set hue."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords: MutableVector, alpha: float) -> Tuple[MutableVector, float]:
        """On color update."""

        if coords[1] < ACHROMATIC_THRESHOLD:
            coords[2] = util.NaN

        return coords, alpha

    @classmethod
    def hue_name(cls) -> str:
        """Hue name."""

        return "hz"

    @classmethod
    def to_base(cls, coords: MutableVector) -> MutableVector:
        """To Jzazbz from JzCzhz."""

        return jzczhz_to_jzazbz(coords)

    @classmethod
    def from_base(cls, coords: MutableVector) -> MutableVector:
        """From Jzazbz to JzCzhz."""

        return jzazbz_to_jzczhz(coords)



"""Calculate range in sRGB."""
import sys
import argparse
import os

sys.path.insert(0, os.getcwd())

from coloraide import Color  # noqa: E402
from coloraide import util  # noqa: E402


def main():
    """Main."""

    parser = argparse.ArgumentParser(
        prog='calc_range_srgb.py', description='Calculate RGB range in the given color.'
    )
    # Flag arguments
    parser.add_argument(
        '--color', '-c', action='store', default='', help="The color whose range relative to RGB will be calculated."
    )
    parser.add_argument(
        '--rgb', '-r', action='store', default='srgb', help="The RGB space which the color will be sized against."
    )
    parser.add_argument(
        '--res', '-s', type=int, default=100, help="Resolution to use when calculating range, default is 100."
    )
    args = parser.parse_args()

    return run(args.color, args.rgb, args.res)


def run(target, rgb, res):
    """Run."""

    max_x = float('-inf')
    max_y = float('-inf')
    max_z = float('-inf')
    min_x = float('inf')
    min_y = float('inf')
    min_z = float('inf')

    print('-> Current:', end="")
    x = y = z = 0
    color = Color(rgb, [0, 0, 0])
    while True:
        color.update(rgb, [x / res, y / res, z / res])
        print('\rCurrent: {}'.format(color.to_string(color=True)), end="")
        cx, cy, cz = color.convert(target).coords()
        if cx < min_x:
            min_x = cx
        if cy < min_y:
            min_y = cy
        if cz < min_z:
            min_z = cz
        if cx > max_x:
            max_x = cx
        if cy > max_y:
            max_y = cy
        if cz > max_z:
            max_z = cz

        if x == y == z == res:
            break
        elif y == z == res:
            x += 1
            y = z = 0
        elif z == res:
            y += 1
            z = 0
        else:
            z += 1

    print('')
    chan_x, chan_y, chan_z = Color('white').convert(target)._space.CHANNEL_NAMES
    print('---- {} range in {} ----'.format(target, rgb))
    print('{}: [{}, {}]'.format(chan_x, util.round_half_up(min_x, 3), util.round_half_up(max_x, 3)))
    print('{}: [{}, {}]'.format(chan_y, util.round_half_up(min_y, 3), util.round_half_up(max_y, 3)))
    print('{}: [{}, {}]'.format(chan_z, util.round_half_up(min_z, 3), util.round_half_up(max_z, 3)))

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Delta E 2000."""
import math
from .. import util
from ..distance import DeltaE
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


class DE2000(DeltaE):
    """Delta E 2000 class."""

    NAME = "2000"

    G_CONST = 25 ** 7

    @classmethod
    def distance(
        cls,
        color: 'Color',
        sample: 'Color',
        kl: float = 1,
        kc: float = 1,
        kh: float = 1,
        **kwargs: Any
    ) -> float:
        """
        Calculate distance doing a direct translation of the algorithm from the CIE Delta E 2000 paper.

        We denoted prime (L') with trailing 'p' and mean is represented with a trailing 'm'.
        Delta has a preceding 'd'. I'm not sure I was completely consistent.

        http://www2.ece.rochester.edu/~gsharma/ciede2000/ciede2000noteCRNA.pdf
        """

        l1, a1, b1 = util.no_nans(color.convert("lab").coords())
        l2, a2, b2 = util.no_nans(sample.convert("lab").coords())

        # Equation (2)
        c1 = math.sqrt(a1 ** 2 + b1 ** 2)
        c2 = math.sqrt(a2 ** 2 + b2 ** 2)

        # Equation (3)
        cm = (c1 + c2) / 2

        # Equation (4)
        c7 = cm ** 7
        g = 0.5 * (1 - math.sqrt(c7 / (c7 + cls.G_CONST)))

        # Equation (5)
        ap1 = (1 + g) * a1
        ap2 = (1 + g) * a2

        # Equation (6)
        cp1 = math.sqrt(ap1 ** 2 + b1 ** 2)
        cp2 = math.sqrt(ap2 ** 2 + b2 ** 2)

        # Equation (7)
        hp1 = 0 if (ap1 == 0 and b1 == 0) else math.atan2(b1, ap1)
        hp2 = 0 if (ap2 == 0 and b2 == 0) else math.atan2(b2, ap2)
        hp1 = math.degrees(hp1 + 2 * math.pi if hp1 < 0.0 else hp1)
        hp2 = math.degrees(hp2 + 2 * math.pi if hp2 < 0.0 else hp2)

        # Equation (8)
        dl = l1 - l2

        # Equation (9)
        dc = cp1 - cp2

        # Equation (10)
        hdiff = hp1 - hp2
        if cp1 * cp2 == 0.0:
            dh = 0.0
        elif abs(hdiff) <= 180.0:
            dh = hdiff
        else:
            # If not `hdiff > 180.0` and not `abs(hdiff) <= 180.0`
            # then it must be `hdiff < -180`
            offset = -360 if hdiff > 180.0 else 360
            dh = hdiff + offset

        # Equation (11)
        dh = 2 * math.sqrt(cp2 * cp1) * math.sin(math.radians(dh / 2))

        # Equation (12)
        lpm = (l1 + l2) / 2

        # Equation (13)
        cpm = (cp1 + cp2) / 2

        # Equation (14)
        hsum = hp1 + hp2
        if cp1 * cp2 == 0:
            hpm = hsum
        elif abs(hp1 - hp2) > 180:
            # if not `hsum < 360`
            # then it must be `hsum >= 360`
            offset = 360 if hsum < 360 else -360
            hpm = (hsum + offset) / 2
        else:  # `abs(hp1 - hp2) <= 180`
            hpm = hsum / 2

        # Equation (15)
        t = (
            1 -
            (0.17 * math.cos(math.radians(hpm - 30))) +
            (0.24 * math.cos(math.radians(2 * hpm))) +
            (0.32 * math.cos(math.radians((3 * hpm) + 6))) -
            (0.20 * math.cos(math.radians((4 * hpm) - 63)))
        )

        # Equation (16)
        dt = 30 * math.exp(-1 * ((hpm - 275) / 25) ** 2)

        # Equation (17)
        cpm7 = cpm ** 7
        rc = 2 * math.sqrt(cpm7 / (cpm7 + cls.G_CONST))

        # Equation (18)
        l_temp = (lpm - 50) ** 2
        sl = 1 + ((0.015 * l_temp) / math.sqrt(20 + l_temp))

        # Equation (19)
        sc = 1 + 0.045 * cpm

        # Equation (20)
        sh = 1 + 0.015 * cpm * t

        # Equation (21)
        rt = -1 * math.sin(math.radians(2 * dt)) * rc

        # Equation (22)
        return math.sqrt(
            (dl / (kl * sl)) ** 2 +
            (dc / (kc * sc)) ** 2 +
            (dh / (kh * sh)) ** 2 +
            rt * (dc / (kc * sc)) * (dh / (kh * sh))
        )



"""Test DIN99o LCH library."""
import unittest
from . import util
from coloraide import Color, NaN


class TestDIN99oLCHInputOutput(util.ColorAsserts, unittest.TestCase):
    """Test DIN99o LCH."""

    def test_input_raw(self):
        """Test raw input."""

        self.assertColorEqual(Color("din99o-lch", [20, 10, 130]), Color('color(--din99o-lch 20% 10 130)'))

    def test_color_class(self):
        """Test raw input."""

        self.assertColorEqual(Color(Color("din99o-lch", [20, 10, 130])), Color('color(--din99o-lch 20% 10 130)'))

    def test_space(self):
        """Test space input and space output format."""

        args = {}

        color = "color(--din99o-lch 20% 10 130)"
        din99o = Color(color)
        self.assertEqual('color(--din99o-lch 20 10 130)', din99o.to_string(**args))

        color = "color(--din99o-lch 20% 10 130 / 1)"
        din99o = Color(color)
        self.assertEqual("color(--din99o-lch 20 10 130)", din99o.to_string(**args))

        color = "color(--din99o-lch 20% 10 130 / 0.2)"
        din99o = Color(color)
        self.assertEqual("color(--din99o-lch 20 10 130 / 0.2)", din99o.to_string(**args))

    def test_percent(self):
        """Test that percents work properly."""

        color = "color(--din99o-lch 20% 10 130 / 100%)"
        din99o = Color(color)
        self.assertEqual("color(--din99o-lch 20 10 130)", din99o.to_string())

        color = "color(--din99o-lch 20% 10 130 / 20%)"
        din99o = Color(color)
        self.assertEqual("color(--din99o-lch 20 10 130 / 0.2)", din99o.to_string())

    def test_no_alpha(self):
        """Test no alpha."""

        color = "color(--din99o-lch 20% 10 130 / 0.2)"
        din99o = Color(color)
        self.assertEqual("color(--din99o-lch 20 10 130)", din99o.to_string(alpha=False))

    def test_force_alpha(self):
        """Test force alpha."""

        color = "color(--din99o-lch 20% 10 130 / 1)"
        din99o = Color(color)
        self.assertEqual("color(--din99o-lch 20 10 130 / 1)", din99o.to_string(alpha=True))

    def test_precision(self):
        """Test precision."""

        color = 'color(--din99o-lch 20.1234567% 10.1234567 130.1234567)'
        self.assertEqual(Color(color).to_string(), 'color(--din99o-lch 20.123 10.123 130.12)')
        self.assertEqual(Color(color).to_string(precision=3), 'color(--din99o-lch 20.1 10.1 130)')
        self.assertEqual(Color(color).to_string(precision=0), 'color(--din99o-lch 20 10 130)')
        self.assertEqual(
            Color(color).to_string(precision=-1),
            'color(--din99o-lch 20.12345669999999842048055143095552921295166015625 10.1234567000000001968373908312059938907623291015625 130.123456699999991315053193829953670501708984375)'  # noqa:  E501
        )

    def test_fit(self):
        """Test fit."""

        self.assertEqual(
            Color('color(--din99o-lch 20% 200 120)').to_string(),
            'color(--din99o-lch 20 200 120)'
        )

        self.assertEqual(
            Color('color(--din99o-lch 20% 200 120)').to_string(fit="clip"),
            'color(--din99o-lch 20 200 120)'
        )

        self.assertEqual(
            Color('color(--din99o-lch 20% 200 120)').to_string(fit=False),
            'color(--din99o-lch 20 200 120)'
        )


class TestDIN99oLCHProperties(util.ColorAsserts, unittest.TestCase):
    """Test DIN99o LCH."""

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--din99o-lch 90% 50 120 / 1)')
        self.assertEqual(c.lightness, 90)
        c.lightness = 80
        self.assertEqual(c.lightness, 80)

    def test_chroma(self):
        """Test `chroma`."""

        c = Color('color(--din99o-lch 90% 50 120 / 1)')
        self.assertEqual(c.chroma, 50)
        c.chroma = 40
        self.assertEqual(c.chroma, 40)

    def test_hue(self):
        """Test `hue`."""

        c = Color('color(--din99o-lch 90% 50 120 / 1)')
        self.assertEqual(c.hue, 120)
        c.hue = 110
        self.assertEqual(c.hue, 110)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--din99o-lch 90% 50 120 / 1)')
        self.assertEqual(c.alpha, 1)
        c.alpha = 0.5
        self.assertEqual(c.alpha, 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('din99o-lch', [90, 50, NaN], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_auto_null(self):
        """Test auto null."""

        c = Color('color(--din99o-lch 90% 0 120 / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_near_zero_null(self):
        """
        Test very near zero null.

        This is a fix up to help give more sane hues
        when chroma is very close to zero.
        """

        c = Color('color(--din99o-lch 90% 0.000000000009 120 / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_from_din99o(self):
        """Test null from DIN99o conversion."""

        c1 = Color('color(--din99o 90% 0 0)')
        c2 = c1.convert('din99o-lch')
        self.assertColorEqual(c2, Color('color(--din99o-lch 90% 0 0)'))
        self.assertTrue(c2.is_nan('hue'))

    def test_achromatic_hue(self):
        """Test that all RGB-ish colors convert to DIN99o LCH with a null hue."""

        for space in ('srgb', 'display-p3', 'rec2020', 'a98-rgb', 'prophoto-rgb'):
            for x in range(0, 256):
                color = Color('color({space} {num:f} {num:f} {num:f})'.format(space=space, num=x / 255))
                color2 = color.convert('din99o-lch')
                self.assertTrue(color2.is_nan('hue'))


class TestQuirks(util.ColorAsserts, unittest.TestCase):
    """Test quirks."""

    def test_chorma_less_than_zero_convert(self):
        """Test handling of negative chroma when converting to lab."""

        self.assertColorEqual(
            Color('color(--din99o-lch 90% -10 120 / 1)').convert('din99o'),
            Color('color(--din99o 90% 0 0)')
        )



"""Test Oklch library."""
import unittest
from . import util
from coloraide import NaN
from coloraide import Color as ColorOrig
from coloraide.spaces.oklch.css import Oklch


class Color(ColorOrig):
    """Utilize Oklab CSS (currently) experimental."""


Color.register(Oklch, overwrite=True)


class TestOklchInputOutput(util.ColorAsserts, unittest.TestCase):
    """Test Oklch."""

    def test_input_raw(self):
        """Test raw input."""

        self.assertColorEqual(Color("oklch", [0.9, 0.5, 270]), Color('oklch(90% 0.5 270)'))

    def test_color_class(self):
        """Test raw input."""

        self.assertColorEqual(Color(Color("oklch", [0.9, 0.5, 270])), Color('oklch(90% 0.5 270)'))

    def test_color(self):
        """Test color input/output format."""

        args = {"color": True}
        color = "color(--oklch 0.9 0.5 270)"

        self.assertEqual(Color(color).to_string(**args), 'color(--oklch 0.9 0.5 270)')

        color = "color(--oklch 0.9 0.5 270 / 0.5)"
        self.assertEqual(Color(color).to_string(**args), 'color(--oklch 0.9 0.5 270 / 0.5)')

        color = "color(--oklch 90% 0.5 270 / 50%)"
        self.assertEqual(Color(color).to_string(**args), 'color(--oklch 0.9 0.5 270 / 0.5)')

    def test_comma(self):
        """Test comma input and comma output format."""

        args = {"comma": True}

        color = "oklch(20%, 0.1, 130)"
        lch = Color(color)
        self.assertEqual(color, lch.to_string(**args))

        color = "oklch(20%, 0.1, 130, 1)"
        lch = Color(color)
        self.assertEqual("oklch(20%, 0.1, 130)", lch.to_string(**args))

        color = "oklch(20%, 0.1, 130, 0.2)"
        lch = Color(color)
        self.assertEqual("oklch(20%, 0.1, 130, 0.2)", lch.to_string(**args))

    def test_space(self):
        """Test space input and space output format."""

        args = {}

        color = "oklch(20% 0.1 130)"
        lch = Color(color)
        self.assertEqual(color, lch.to_string(**args))

        color = "oklch(20% 0.1 130 / 1)"
        lch = Color(color)
        self.assertEqual("oklch(20% 0.1 130)", lch.to_string(**args))

        color = "oklch(20% 0.1 130 / 0.2)"
        lch = Color(color)
        self.assertEqual(color, lch.to_string(**args))

    def test_percent(self):
        """Test that percents work properly."""

        args = {"comma": True}

        color = "oklch(20%, 0.1, 130, 100%)"
        lch = Color(color)
        self.assertEqual("oklch(20%, 0.1, 130)", lch.to_string(**args))

        color = "oklch(20%, 0.1, 130, 20%)"
        lch = Color(color)
        self.assertEqual("oklch(20%, 0.1, 130, 0.2)", lch.to_string(**args))

        args["comma"] = False

        color = "oklch(20% 0.1 130 / 100%)"
        lch = Color(color)
        self.assertEqual("oklch(20% 0.1 130)", lch.to_string(**args))

        color = "oklch(20% 0.1 130 / 20%)"
        lch = Color(color)
        self.assertEqual("oklch(20% 0.1 130 / 0.2)", lch.to_string(**args))

    def test_no_alpha(self):
        """Test no alpha."""

        args = {"alpha": False}

        color = "oklch(90% 0.5 270 / 0.5)"
        oklch = Color(color)
        self.assertEqual("oklch(90% 0.5 270)", oklch.to_string(**args))

    def test_force_alpha(self):
        """Test force alpha."""

        args = {"alpha": True}

        color = "oklch(90% 0.5 270 / 100%)"
        oklch = Color(color)
        self.assertEqual("oklch(90% 0.5 270 / 1)", oklch.to_string(**args))

    def test_precision(self):
        """Test precision."""

        color = 'color(--oklch 0.123456 0.123456 0.123456)'
        self.assertEqual(Color(color).to_string(), 'oklch(12.346% 0.12346 0.12346)')
        self.assertEqual(Color(color).to_string(precision=3), 'oklch(12.3% 0.123 0.123)')
        self.assertEqual(Color(color).to_string(precision=0), 'oklch(12% 0 0)')
        self.assertEqual(
            Color(color).to_string(precision=-1),
            'oklch(12.3455999999999992411403582082130014896392822265625% 0.12345599999999999629718416827017790637910366058349609 0.12345599999999999629718416827017790637910366058349609)'  # noqa:  E501
        )

    def test_fit(self):
        """Test fit."""

        self.assertEqual(
            Color('oklch(90% 200 270)').to_string(),
            'oklch(90% 200 270)'
        )

        self.assertEqual(
            Color('oklch(90% 200 270)').to_string(fit="clip"),
            'oklch(90% 200 270)'
        )

        self.assertEqual(
            Color('oklch(90% 200 270)').to_string(fit=False),
            'oklch(90% 200 270)'
        )


class TestOklchProperties(util.ColorAsserts, unittest.TestCase):
    """Test Oklch."""

    def test_lightness(self):
        """Test `lightness`."""

        c = Color('color(--oklch 0.9 0.5 270 / 1)')
        self.assertEqual(c.lightness, 0.9)
        c.lightness = 0.2
        self.assertEqual(c.lightness, 0.2)

    def test_chroma(self):
        """Test `chroma`."""

        c = Color('color(--oklch 0.9 0.5 270 / 1)')
        self.assertEqual(c.chroma, 0.5)
        c.chroma = 0.1
        self.assertEqual(c.chroma, 0.1)

    def test_hue(self):
        """Test `hue`."""

        c = Color('color(--oklch 0.9 0.5 270 / 1)')
        self.assertEqual(c.hue, 270)
        c.hue = 0.1
        self.assertEqual(c.hue, 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--oklch 0.9 0.5 270 / 1)')
        self.assertEqual(c.alpha, 1)
        c.alpha = 0.5
        self.assertEqual(c.alpha, 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('oklch', [0.9, 50, NaN], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_auto_null(self):
        """Test auto null."""

        c = Color('color(--oklch 90% 0 120 / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_near_zero_null(self):
        """
        Test very near zero null.

        This is a fix up to help give more sane hues
        when chroma is very close to zero.
        """

        c = Color('color(--oklch 90% 0.0000009 120 / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_from_oklab(self):
        """Test null from Lab conversion."""

        c1 = Color('color(--oklab 90% 0 0)')
        c2 = c1.convert('oklch')
        self.assertColorEqual(c2, Color('color(--oklch 90% 0 0)'))
        self.assertTrue(c2.is_nan('hue'))

    def test_achromatic_hue(self):
        """Test that all RGB-ish colors convert to Oklch with a null hue."""

        for space in ('srgb', 'display-p3', 'rec2020', 'a98-rgb', 'prophoto-rgb'):
            for x in range(0, 256):
                color = Color('color({space} {num:f} {num:f} {num:f})'.format(space=space, num=x / 255))
                color2 = color.convert('oklch')
                self.assertTrue(color2.is_nan('hue'))


class TestQuirks(util.ColorAsserts, unittest.TestCase):
    """Test quirks."""

    def test_chorma_less_than_zero_convert(self):
        """Test handling of negative chroma when converting to Oklab."""

        self.assertColorEqual(Color('color(--oklch 90% -10 120 / 1)').convert('oklab'), Color('color(--oklab 0.9 0 0)'))



"""Sanity check that ensures all colors round trip back."""
from coloraide import Color
import pytest

SPACES = Color.CS_MAP.keys()


class TestRoundTrip:
    """
    Test round trip conversions of all color spaces.

    A color that does not round trip back likely has a broken conversion.
    It may be possible that some color spaces degrade through conversion,
    but the currently supported spaces have enough integrity when performing
    a round trip to meet the required value at the default precision.

    There are certain values they may not round trip exactly. Hues may become
    null, etc. But we are picking general colors that should round trip well
    enough.
    """

    # Skip colors with null hues or hues that can wrap.
    COLORS = [
        Color('red'),
        Color('orange'),
        Color('yellow'),
        Color('green'),
        Color('blue'),
        Color('indigo'),
        Color('violet'),
        # Color('white'),  # General: this is difficult as not everything lands exactly
        Color('black')
    ]

    def assert_round_trip(self, color, space):
        """Cycle through all the other colors and convert to them and back and check the results."""

        c1 = Color(color).convert(space)
        for space in SPACES:
            # Print the color space to easily identify which color space broke.
            c2 = c1.convert(space)
            c2.convert(c1.space(), in_place=True)
            str1 = c1.to_string(color=True)
            # Run back through parsing in case we hit something like a hue that needs normalization.
            str2 = Color(c2.to_string(color=True)).to_string(color=True)
            if str1 != str2:
                print('----- Convert: {} <=> {} -----'.format(c1.space(), space))
                print('Original: ', color.to_string(color=True))
                print(c1.space() + ': ', str1, c1.coords())
                print(space + ': ', str2, c2.coords())
                assert str1 == str2

    @pytest.mark.parametrize('space', SPACES)
    def test_round_trip(self, space):
        """Test round trip."""

        for c in self.COLORS:
            self.assert_round_trip(c, space)



"""Calculate XYZ conversion matrices."""
import numpy as np

np.set_printoptions(precision=None, sign='-', floatmode='unique')


def xy_to_xyz(x, y):
    """Convert `xyY` to `xyz`."""

    return [x / y, 1, (1 - x - y) / y]


white_d65 = np.asfarray(xy_to_xyz(0.31270, 0.32900))
white_d50 = np.asfarray(xy_to_xyz(0.34570, 0.35850))


def get_matrix(wp, space):
    """Get the matrices for the specified space."""

    if space == 'srgb':
        x = np.asfarray([0.64, 0.30, 0.15])
        y = np.asfarray([0.33, 0.60, 0.06])
    elif space == 'display-p3':
        x = np.asfarray([0.68, 0.265, 0.150])
        y = np.asfarray([0.32, 0.69, 0.060])
    elif space == 'rec2020':
        x = np.asfarray([0.708, 0.17, 0.131])
        y = np.asfarray([0.292, 0.797, 0.046])
    elif space == 'a98-rgb':
        x = np.asfarray([0.64, 0.21, 0.15])
        y = np.asfarray([0.33, 0.71, 0.06])
    elif space == 'prophoto-rgb':
        x = np.asfarray([0.7347, 0.1596, 0.0366])
        y = np.asfarray([0.2653, 0.8404, 0.0001])
    else:
        raise ValueError

    one = np.float64(1.0)
    m = np.asfarray(
        [
            [x[0] / y[0], one, (one - x[0] - y[0]) / y[0]],
            [x[1] / y[1], one, (one - x[1] - y[1]) / y[1]],
            [x[2] / y[2], one, (one - x[2] - y[2]) / y[2]]
        ]
    )
    mi = np.linalg.inv(m)

    rgb = np.dot(wp, mi).reshape(3, 1)
    rgb2xyz = np.multiply(rgb, m).transpose()
    xyz2rgb = np.linalg.inv(rgb2xyz)

    return rgb2xyz, xyz2rgb


if __name__ == "__main__":
    print('===== sRGB =====')
    to_xyz, from_xyz = get_matrix(white_d65, 'srgb')
    print('--- rgb -> xyz ---')
    print(to_xyz)
    print('--- xyz -> rgb ---')
    print(from_xyz)

    print('===== Display P3 =====')
    to_xyz, from_xyz = get_matrix(white_d65, 'display-p3')
    print('--- rgb -> xyz ---')
    print(to_xyz)
    print('--- xyz -> rgb ---')
    print(from_xyz)

    print('===== Adobe 98 =====')
    to_xyz, from_xyz = get_matrix(white_d65, 'a98-rgb')
    print('--- rgb -> xyz ---')
    print(to_xyz)
    print('--- xyz -> rgb ---')
    print(from_xyz)

    print('===== Rec.2020 =====')
    to_xyz, from_xyz = get_matrix(white_d65, 'rec2020')
    print('--- rgb -> xyz ---')
    print(to_xyz)
    print('--- xyz -> rgb ---')
    print(from_xyz)

    print('===== ProPhoto =====')
    to_xyz, from_xyz = get_matrix(white_d50, 'prophoto-rgb')
    print('--- rgb -> xyz ---')
    print(to_xyz)
    print('--- xyz -> rgb ---')
    print(from_xyz)



"""Test ProPhoto RGB library."""
import unittest
from . import util
from coloraide import Color


class TestProPhotoRGBInputOutput(util.ColorAsserts, unittest.TestCase):
    """Test ProPhoto RGB."""

    def test_input_raw(self):
        """Test raw input."""

        self.assertColorEqual(Color("prophoto-rgb", [1, 1, 1]), Color('color(prophoto-rgb 100% 100% 100%)'))

    def test_color_class(self):
        """Test raw input."""

        self.assertColorEqual(Color(Color("prophoto-rgb", [1, 1, 1])), Color('color(prophoto-rgb 100% 100% 100%)'))

    def test_color(self):
        """Test color input/output format."""

        args = {"color": True}
        color = "color(prophoto-rgb 0.3 1 0.5)"

        self.assertEqual(Color(color).to_string(**args), 'color(prophoto-rgb 0.3 1 0.5)')

        color = "color(prophoto-rgb 0.3 1 0.5 / 0.5)"
        self.assertEqual(Color(color).to_string(**args), 'color(prophoto-rgb 0.3 1 0.5 / 0.5)')

        color = "color(prophoto-rgb 30% 100% 50% / 50%)"
        self.assertEqual(Color(color).to_string(**args), 'color(prophoto-rgb 0.3 1 0.5 / 0.5)')

    def test_no_alpha(self):
        """Test no alpha."""

        args = {"alpha": False}

        color = "color(prophoto-rgb 0.3 1 0.5 / 0.2)"
        pro = Color(color)
        self.assertEqual("color(prophoto-rgb 0.3 1 0.5)", pro.to_string(**args))

    def test_force_alpha(self):
        """Test force alpha."""

        args = {"alpha": True}

        color = "color(prophoto-rgb 0.3 1 0.5 / 100%)"
        pro = Color(color)
        self.assertEqual("color(prophoto-rgb 0.3 1 0.5 / 1)", pro.to_string(**args))

    def test_precision(self):
        """Test precision."""

        color = 'color(prophoto-rgb 0.1234567 0.1234567 0.1234567)'
        self.assertEqual(Color(color).to_string(), 'color(prophoto-rgb 0.12346 0.12346 0.12346)')
        self.assertEqual(Color(color).to_string(precision=3), 'color(prophoto-rgb 0.123 0.123 0.123)')
        self.assertEqual(Color(color).to_string(precision=0), 'color(prophoto-rgb 0 0 0)')
        self.assertEqual(
            Color(color).to_string(precision=-1),
            'color(prophoto-rgb 0.12345670000000000254836152180359931662678718566894531 0.12345670000000000254836152180359931662678718566894531 0.12345670000000000254836152180359931662678718566894531)'  # noqa:  E501
        )

    def test_fit(self):
        """Test fit."""

        self.assertEqual(
            Color('color(prophoto-rgb 2 -1 0)').to_string(),
            'color(prophoto-rgb 1 0.38138 0.55612)'
        )

        self.assertEqual(
            Color('color(prophoto-rgb 2 -1 0)').to_string(fit="clip"),
            'color(prophoto-rgb 1 0 0)'
        )

        self.assertEqual(
            Color('color(prophoto-rgb 2 -1 0)').to_string(fit=False),
            'color(prophoto-rgb 2 -1 0)'
        )


class TestProPhotoRGBProperties(util.ColorAsserts, unittest.TestCase):
    """Test ProPhoto RGB."""

    def test_red(self):
        """Test `red`."""

        c = Color('color(prophoto-rgb 0.1 0.2 0.3 / 1)')
        self.assertEqual(c.red, 0.1)
        c.red = 0.2
        self.assertEqual(c.red, 0.2)

    def test_green(self):
        """Test `green`."""

        c = Color('color(prophoto-rgb 0.1 0.2 0.3 / 1)')
        self.assertEqual(c.green, 0.2)
        c.green = 0.1
        self.assertEqual(c.green, 0.1)

    def test_blue(self):
        """Test `blue`."""

        c = Color('color(prophoto-rgb 0.1 0.2 0.3 / 1)')
        self.assertEqual(c.blue, 0.3)
        c.blue = 0.1
        self.assertEqual(c.blue, 0.1)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(prophoto-rgb 0.1 0.2 0.3 / 1)')
        self.assertEqual(c.alpha, 1)
        c.alpha = 0.5
        self.assertEqual(c.alpha, 0.5)



"""
LCH class.

---- License ----

Copyright (c) 2021 Björn Ottosson

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from ...spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, Lchish, FLG_ANGLE, FLG_OPT_PERCENT
from ... import util
import re
import math
from ...util import Vector, MutableVector
from typing import Tuple

ACHROMATIC_THRESHOLD = 0.000002


def oklab_to_oklch(oklab: Vector) -> MutableVector:
    """Oklab to Oklch."""

    l, a, b = oklab

    c = math.sqrt(a ** 2 + b ** 2)
    h = math.degrees(math.atan2(b, a))

    # Achromatic colors will often get extremely close, but not quite hit zero.
    # Essentially, we want to discard noise through rounding and such.
    if c < ACHROMATIC_THRESHOLD:
        h = util.NaN

    return [l, c, util.constrain_hue(h)]


def oklch_to_oklab(oklch: Vector) -> MutableVector:
    """Oklch to Oklab."""

    l, c, h = oklch
    h = util.no_nan(h)

    # If, for whatever reason (mainly direct user input),
    # if chroma is less than zero, clamp to zero.
    if c < 0.0:
        c = 0.0

    return [
        l,
        c * math.cos(math.radians(h)),
        c * math.sin(math.radians(h))
    ]


class Oklch(Lchish, Space):
    """Oklch class."""

    BASE = "oklab"
    NAME = "oklch"
    SERIALIZE = ("--oklch",)
    CHANNEL_NAMES = ("l", "c", "h")
    CHANNEL_ALIASES = {
        "lightness": "l",
        "chroma": "c",
        "hue": "h"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"

    BOUNDS = (
        GamutUnbound(0.0, 1.0, FLG_OPT_PERCENT),
        GamutUnbound(0.0, 1.0),
        GamutUnbound(0.0, 360.0, FLG_ANGLE)
    )

    @property
    def l(self) -> float:
        """Lightness."""

        return self._coords[0]

    @l.setter
    def l(self, value: float) -> None:
        """Get true luminance."""

        self._coords[0] = self._handle_input(value)

    @property
    def c(self) -> float:
        """Chroma."""

        return self._coords[1]

    @c.setter
    def c(self, value: float) -> None:
        """chroma."""

        self._coords[1] = self._handle_input(value)

    @property
    def h(self) -> float:
        """Hue."""

        return self._coords[2]

    @h.setter
    def h(self, value: float) -> None:
        """Shift the hue."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords: MutableVector, alpha: float) -> Tuple[MutableVector, float]:
        """On color update."""

        if coords[1] < ACHROMATIC_THRESHOLD:
            coords[2] = util.NaN

        return coords, alpha

    @classmethod
    def to_base(cls, oklch: Vector) -> MutableVector:
        """To Lab."""

        return oklch_to_oklab(oklch)

    @classmethod
    def from_base(cls, oklab: Vector) -> MutableVector:
        """To Lab."""

        return oklab_to_oklch(oklab)



"""Lch D65 class."""
from ..spaces import RE_DEFAULT_MATCH
from .lch import Lch
import re


class LchD65(Lch):
    """Lch D65 class."""

    BASE = "lab-d65"
    NAME = "lch-d65"
    SERIALIZE = ("--lch-d65",)
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"



"""SRGB color class."""
import re
from . import color_names
from .. import srgb as base
from ... import parse
from ... import util
from typing import Optional, Union, Any, Tuple, TYPE_CHECKING
from ...util import MutableVector

if TYPE_CHECKING:  # pragma: no cover
    from ...color import Color

RE_COMPRESS = re.compile(r'(?i)^#({hex})\1({hex})\2({hex})\3(?:({hex})\4)?$'.format(**parse.COLOR_PARTS))


class SRGB(base.SRGB):
    """SRGB class."""

    MATCH = re.compile(
        r"""(?xi)
        (?:
            # RGB syntax
            \brgba?\(\s*
            (?:
                # Space separated format
                (?:
                    # Float form
                    (?:{float}{space}){{2}}{float} |
                    # Percent form
                    (?:{percent}{space}){{2}}{percent}
                )({slash}(?:{percent}|{float}))? |
                # Comma separated format
                (?:
                    # Float form
                    (?:{float}{comma}){{2}}{float} |
                    # Percent form
                    (?:{percent}{comma}){{2}}{percent}
                )({comma}(?:{percent}|{float}))?
            )
            \s*\) |
            # Hex syntax
            \#(?:{hex}{{6}}(?:{hex}{{2}})?|{hex}{{3}}(?:{hex})?)\b |
            # Names
            \b(?<!\#)[a-z]{{3,}}(?!\()\b
        )
        """.format(**parse.COLOR_PARTS)
    )

    def to_string(
        self,
        parent: 'Color',
        *,
        alpha: Optional[bool] = None,
        precision: Optional[int] = None,
        fit: Union[bool, str] = True,
        none: bool = False,
        **kwargs: Any
    ) -> str:
        """Convert to CSS."""

        if precision is None:
            precision = parent.PRECISION

        options = kwargs
        if options.get("color"):
            return super().to_string(parent, alpha=alpha, precision=precision, fit=fit, none=none, **kwargs)

        a = util.no_nan(self.alpha) if not none else self.alpha
        alpha = alpha is not False and (alpha is True or a < 1.0 or util.is_nan(a))
        compress = options.get("compress", False)

        # Handle hex and color names
        value = ''
        if options.get("hex") or options.get("names"):
            h = self._get_hex(parent, upper=options.get("upper", False), alpha=alpha, fit=fit)
            if options.get("hex"):
                value = h
                if compress:
                    m = RE_COMPRESS.match(value)
                    if m:
                        value = m.expand(r"#\1\2\3\4") if alpha else m.expand(r"#\1\2\3")
            if options.get("names"):
                length = len(h) - 1
                index = int(length / 4)
                if length in (8, 4) and h[-index:].lower() == ("f" * index):
                    h = h[:-index]
                n = color_names.hex2name(h)
                if n is not None:
                    value = n

        # Handle normal RGB function format.
        if not value:
            percent = options.get("percent", False)
            comma = options.get("comma", False)
            factor = 100.0 if percent else 255.0

            method = None if not isinstance(fit, str) else fit
            coords = parent.fit(method=method).coords() if fit else self.coords()
            if not none:
                coords = util.no_nans(coords)

            fmt = util.fmt_percent if percent else util.fmt_float
            if alpha:
                template = "rgba({}, {}, {}, {})" if comma else "rgb({} {} {} / {})"
                value = template.format(
                    fmt(coords[0] * factor, precision),
                    fmt(coords[1] * factor, precision),
                    fmt(coords[2] * factor, precision),
                    util.fmt_float(a, max(util.DEF_PREC, precision))
                )
            else:
                template = "rgb({}, {}, {})" if comma else "rgb({} {} {})"
                value = template.format(
                    fmt(coords[0] * factor, precision),
                    fmt(coords[1] * factor, precision),
                    fmt(coords[2] * factor, precision),
                )

        return value

    def _get_hex(
        self,
        parent: 'Color',
        *,
        upper: bool = False,
        alpha: bool = False,
        fit: Union[str, bool] = True
    ) -> str:
        """Get the hex `RGB` value."""

        method = None if not isinstance(fit, str) else fit
        coords = util.no_nans(parent.fit(method=method).coords())

        template = "#{:02x}{:02x}{:02x}{:02x}" if alpha else "#{:02x}{:02x}{:02x}"
        if upper:
            template = template.upper()

        if alpha:
            value = template.format(
                int(util.round_half_up(coords[0] * 255.0)),
                int(util.round_half_up(coords[1] * 255.0)),
                int(util.round_half_up(coords[2] * 255.0)),
                int(util.round_half_up(util.no_nan(self.alpha) * 255.0))
            )
        else:
            value = template.format(
                int(util.round_half_up(coords[0] * 255.0)),
                int(util.round_half_up(coords[1] * 255.0)),
                int(util.round_half_up(coords[2] * 255.0))
            )
        return value

    @classmethod
    def translate_channel(cls, channel: int, value: str) -> float:
        """Translate channel string."""

        if channel in (0, 1, 2):
            if value.startswith('#'):
                return parse.norm_hex_channel(value)
            else:
                return parse.norm_rgb_channel(value)
        elif channel == -1:
            if value.startswith('#'):
                return parse.norm_hex_channel(value)
            else:
                return parse.norm_alpha_channel(value)
        else:  # pragma: no cover
            raise ValueError('{} is not a valid channel index'.format(channel))

    @classmethod
    def split_channels(cls, color: str) -> Tuple[MutableVector, float]:
        """Split channels."""

        if color[:3].lower().startswith('rgb'):
            start = 5 if color[:4].lower().startswith('rgba') else 4
            channels = []
            alpha = 1.0
            for i, c in enumerate(parse.RE_CHAN_SPLIT.split(color[start:-1].strip()), 0):
                c = c.lower()
                if i <= 2:
                    channels.append(cls.translate_channel(i, c))
                elif i == 3:
                    alpha = cls.translate_channel(-1, c)
            return cls.null_adjust(channels, alpha)
        else:
            length = len(color)
            if length in (7, 9):
                return cls.null_adjust(
                    [
                        cls.translate_channel(0, "#" + color[1:3]),
                        cls.translate_channel(1, "#" + color[3:5]),
                        cls.translate_channel(2, "#" + color[5:7])
                    ],
                    cls.translate_channel(-1, "#" + color[7:9]) if length == 9 else 1.0
                )
            else:
                return cls.null_adjust(
                    [
                        cls.translate_channel(0, "#" + color[1] * 2),
                        cls.translate_channel(1, "#" + color[2] * 2),
                        cls.translate_channel(2, "#" + color[3] * 2)
                    ],
                    cls.translate_channel(-1, "#" + color[4] * 2) if length == 5 else 1.0
                )

    @classmethod
    def match(
        cls,
        string: str,
        start: int = 0,
        fullmatch: bool = True
    ) -> Optional[Tuple[Tuple[MutableVector, float], int]]:
        """Match a CSS color string."""

        match = super().match(string, start, fullmatch)
        if match is not None:
            return match
        m = cls.MATCH.match(string, start)
        if m is not None and (not fullmatch or m.end(0) == len(string)):
            string = string[m.start(0):m.end(0)].lower()
            if not string.startswith(('#', 'rgb')):
                value = color_names.name2hex(string)
                if value is not None:
                    return cls.split_channels(value), m.end(0)
            else:
                return cls.split_channels(string), m.end(0)

        return None



"""
Calculate `oklab` matrices.

https://github.com/w3c/csswg-drafts/issues/6642#issuecomment-945714988
"""
import numpy as np
import sys
import os

sys.path.insert(0, os.getcwd())

import tools.calc_xyz_transform as xyzt  # noqa: E402

np.set_printoptions(precision=None, sign='-', floatmode='unique')

# Calculated using our own `calc_xyz_transform.py`
RGB_TO_XYZ, XYZ_TO_RGB = xyzt.get_matrix(xyzt.white_d65, 'srgb')

M0 = np.asfarray(
    [
        [0.77849780, 0.34399940, -0.12249720],
        [0.03303601, 0.93076195, 0.03620204],
        [0.05092917, 0.27933344, 0.66973739]
    ]
)

XYZ_TO_LMS = np.divide(M0, np.outer(M0.dot(xyzt.white_d65), np.ones(3)))
LMS_TO_XYZ = np.linalg.inv(XYZ_TO_LMS)

SRGBL_TO_LMS = np.dot(XYZ_TO_LMS, RGB_TO_XYZ)
LMS_TO_SRGBL = np.linalg.inv(SRGBL_TO_LMS)


LMS3_TO_OKLAB = np.asfarray(
    [
        [0.2104542553, 0.793617785, -0.0040720468],
        [1.9779984951, -2.428592205, 0.4505937099],
        [0.0259040371, 0.7827717662, -0.808675766]
    ]
)

OKLAB_TO_LMS3 = np.linalg.inv(LMS3_TO_OKLAB)

if __name__ == "__main__":
    print('===== sRGB Linear -> lms =====')
    print(SRGBL_TO_LMS)
    print('===== lms -> sRGB Linear =====')
    print(LMS_TO_SRGBL)
    print('===== XYZ D65 Linear -> lms =====')
    print(XYZ_TO_LMS)
    print('===== lms -> XYZ D65 =====')
    print(LMS_TO_XYZ)
    print('===== lms ** 1/3 -> Oklab =====')
    print(LMS3_TO_OKLAB)
    print('===== Oklab -> lms ** 1/3 =====')
    print(OKLAB_TO_LMS3)



"""
Okhsv class.

Adapted to ColorAide Python and ColorAide by Isaac Muse (2021)

---- License ----

Copyright (c) 2021 Björn Ottosson

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from ..spaces import Space, RE_DEFAULT_MATCH, FLG_ANGLE, FLG_OPT_PERCENT, GamutBound, Cylindrical
from .. import util
from .oklab import oklab_to_linear_srgb
from .okhsl import toe, toe_inv, find_cusp, to_st
import re
import math
from ..util import MutableVector
from typing import Tuple


def okhsv_to_oklab(hsv: MutableVector) -> MutableVector:
    """Convert from Okhsv to Oklab."""

    h, s, v = hsv
    h = util.no_nan(h) / 360.0

    l = toe_inv(v)
    a = b = 0.0

    if l != 0 and s != 0:
        a_ = math.cos(2.0 * math.pi * h)
        b_ = math.sin(2.0 * math.pi * h)

        cusp = find_cusp(a_, b_)
        s_max, t_max = to_st(cusp)
        s_0 = 0.5
        k = 1 - s_0 / s_max

        # first we compute L and V as if the gamut is a perfect triangle:

        # L, C when v==1:
        l_v = 1 - s * s_0 / (s_0 + t_max - t_max * k * s)
        c_v = s * t_max * s_0 / (s_0 + t_max - t_max * k * s)

        l = v * l_v
        c = v * c_v

        # then we compensate for both toe and the curved top part of the triangle:
        l_vt = toe_inv(l_v)
        c_vt = c_v * l_vt / l_v

        l_new = toe_inv(l)
        c = c * l_new / l
        l = l_new

        # RGB scale
        rs, gs, bs = oklab_to_linear_srgb([l_vt, a_ * c_vt, b_ * c_vt])
        scale_l = util.nth_root(1.0 / max(max(rs, gs), max(bs, 0.0)), 3)

        l = l * scale_l
        c = c * scale_l

        a = c * a_
        b = c * b_

    return [l, a, b]


def oklab_to_okhsv(lab: MutableVector) -> MutableVector:
    """Oklab to Okhsv."""

    c = math.sqrt(lab[1] ** 2 + lab[2] ** 2)
    l = lab[0]

    h = util.NaN
    s = 0.0
    v = toe(l)

    if c != 0 and l != 0 and l != 1:
        a_ = lab[1] / c
        b_ = lab[2] / c

        h = 0.5 + 0.5 * math.atan2(-lab[2], -lab[1]) / math.pi

        cusp = find_cusp(a_, b_)
        s_max, t_max = to_st(cusp)
        s_0 = 0.5
        k = 1 - s_0 / s_max

        # first we find `L_v`, `C_v`, `L_vt` and `C_vt`
        t = t_max / (c + l * t_max)
        l_v = t * l
        c_v = t * c

        l_vt = toe_inv(l_v)
        c_vt = c_v * l_vt / l_v

        # we can then use these to invert the step that compensates for the toe and the curved top part of the triangle:
        rs, gs, bs = oklab_to_linear_srgb([l_vt, a_ * c_vt, b_ * c_vt])
        scale_l = util.nth_root(1.0 / max(max(rs, gs), max(bs, 0.0)), 3)

        l = l / scale_l
        c = c / scale_l

        c = c * toe(l) / l
        l = toe(l)

        # we can now compute v and s:
        v = l / l_v
        s = (s_0 + t_max) * c_v / ((t_max * s_0) + t_max * k * c_v)

    if s == 0:
        h = util.NaN

    return [util.constrain_hue(h * 360), s, v]


class Okhsv(Cylindrical, Space):
    """Okhsv class."""

    BASE = "oklab"
    NAME = "okhsv"
    SERIALIZE = ("--okhsv",)
    CHANNEL_NAMES = ("h", "s", "v")
    CHANNEL_ALIASES = {
        "hue": "h",
        "saturation": "s",
        "value": "v"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"
    GAMUT_CHECK = "srgb"

    BOUNDS = (
        GamutBound(0.0, 360.0, FLG_ANGLE),
        GamutBound(0.0, 1.0, FLG_OPT_PERCENT),
        GamutBound(0.0, 1.0, FLG_OPT_PERCENT)
    )

    @property
    def h(self) -> float:
        """Hue channel."""

        return self._coords[0]

    @h.setter
    def h(self, value: float) -> None:
        """Shift the hue."""

        self._coords[0] = self._handle_input(value)

    @property
    def s(self) -> float:
        """Saturation channel."""

        return self._coords[1]

    @s.setter
    def s(self, value: float) -> None:
        """Saturate or unsaturate the color by the given factor."""

        self._coords[1] = self._handle_input(value)

    @property
    def v(self) -> float:
        """Value channel."""

        return self._coords[2]

    @v.setter
    def v(self, value: float) -> None:
        """Set value channel."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords: MutableVector, alpha: float) -> Tuple[MutableVector, float]:
        """On color update."""

        if coords[1] == 0:
            coords[0] = util.NaN
        return coords, alpha

    @classmethod
    def to_base(cls, okhsv: MutableVector) -> MutableVector:
        """To Oklab from Okhsv."""

        return okhsv_to_oklab(okhsv)

    @classmethod
    def from_base(cls, oklab: MutableVector) -> MutableVector:
        """From Oklab to Okhsv."""

        return oklab_to_okhsv(oklab)



"""Gamut handling."""
from .. import util
from ..util import MutableVector
from ..spaces import FLG_ANGLE, GamutBound
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Optional, Any

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


def clip_channels(color: 'Color') -> MutableVector:
    """Clip channels."""

    channels = util.no_nans(color.coords())
    fit = []

    for i, value in enumerate(channels):
        bounds = color._space.BOUNDS[i]
        a = bounds.lower  # type: Optional[float]
        b = bounds.upper  # type: Optional[float]
        is_bound = isinstance(bounds, GamutBound)

        # Wrap the angle. Not technically out of gamut, but we will clean it up.
        if bounds.flags & FLG_ANGLE:
            fit.append(value % 360.0)
            continue

        # These parameters are unbounded
        if not is_bound:  # pragma: no cover
            # Will not execute unless we have a space that defines some coordinates
            # as bound and others as not. We do not currently have such spaces.
            a = b = None

        # Fit value in bounds.
        fit.append(util.clamp(value, a, b))
    return fit


def verify(color: 'Color', tolerance: float) -> bool:
    """Verify the values are in bound."""

    channels = util.no_nans(color.coords())
    for i, value in enumerate(channels):
        bounds = color._space.BOUNDS[i]
        a = bounds.lower  # type: Optional[float]
        b = bounds.upper  # type: Optional[float]
        is_bound = isinstance(bounds, GamutBound)

        # Angles will wrap, so no sense checking them
        if bounds.flags & FLG_ANGLE:
            continue

        # These parameters are unbounded
        if not is_bound:
            a = b = None

        # Check if bounded values are in bounds
        if (a is not None and value < (a - tolerance)) or (b is not None and value > (b + tolerance)):
            return False
    return True


class Fit(ABCMeta):
    """Fit plugin class."""

    NAME = ''

    @classmethod
    @abstractmethod
    def fit(cls, color: 'Color', **kwargs: Any) -> MutableVector:
        """Get coordinates of the new gamut mapped color."""



"""Distance and Delta E."""
from abc import ABCMeta, abstractmethod
from .. import util
import math
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:  # pragma: no cover
    from ..color import Color


def distance_euclidean(color: 'Color', sample: 'Color', space: str = "lab") -> float:
    """
    Euclidean distance.

    https://en.wikipedia.org/wiki/Euclidean_distance
    """

    coords1 = util.no_nans(color.convert(space).coords())
    coords2 = util.no_nans(sample.convert(space).coords())

    return math.sqrt(sum((x - y) ** 2.0 for x, y in zip(coords1, coords2)))


class DeltaE(ABCMeta):
    """Delta E plugin class."""

    NAME = ''

    @classmethod
    @abstractmethod
    def distance(cls, color: 'Color', sample: 'Color', **kwargs: Any) -> float:
        """Get distance between color and sample."""



"""Oklch class."""
import re
from .. import oklch as base
from ... import parse
from ... import util
from ...util import MutableVector
from typing import Union, Optional, Tuple, Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ...color import Color


class Oklch(base.Oklch):
    """Oklch class."""

    MATCH = re.compile(
        r"""(?xi)
        \boklch\(\s*
        (?:
            # Space separated format
            {percent}{space}{float}{space}{angle}(?:{slash}(?:{percent}|{float}))? |
            # comma separated format
            {percent}{comma}{float}{comma}{angle}(?:{comma}(?:{percent}|{float}))?
        )
        \s*\)
        """.format(**parse.COLOR_PARTS)
    )

    def to_string(
        self,
        parent: 'Color',
        *,
        alpha: Optional[bool] = None,
        precision: Optional[int] = None,
        fit: Union[str, bool] = True,
        none: bool = False,
        **kwargs: Any
    ) -> str:
        """Convert to CSS."""

        if precision is None:
            precision = parent.PRECISION

        options = kwargs
        if options.get("color"):
            return super().to_string(parent, alpha=alpha, precision=precision, fit=fit, none=none, **kwargs)

        a = util.no_nan(self.alpha) if not none else self.alpha
        alpha = alpha is not False and (alpha is True or a < 1.0 or util.is_nan(a))
        method = None if not isinstance(fit, str) else fit
        coords = parent.fit(method=method).coords() if fit else self.coords()
        if not none:
            coords = util.no_nans(coords)

        if alpha:
            template = "oklch({}, {}, {}, {})" if options.get("comma") else "oklch({} {} {} / {})"
            return template.format(
                util.fmt_percent(coords[0] * 100, precision),
                util.fmt_float(coords[1], precision),
                util.fmt_float(coords[2], precision),
                util.fmt_float(a, max(util.DEF_PREC, precision))
            )
        else:
            template = "oklch({}, {}, {})" if options.get("comma") else "oklch({} {} {})"
            return template.format(
                util.fmt_percent(coords[0] * 100, precision),
                util.fmt_float(coords[1], precision),
                util.fmt_float(coords[2], precision)
            )

    @classmethod
    def translate_channel(cls, channel: int, value: str) -> float:
        """Translate channel string."""

        if channel == 0:
            return parse.norm_percent_channel(value) / 100
        elif channel == 1:
            return parse.norm_float(value)
        elif channel == 2:
            return parse.norm_angle_channel(value)
        elif channel == -1:
            return parse.norm_alpha_channel(value)
        else:  # pragma: no cover
            raise ValueError('{} is not a valid channel index'.format(channel))

    @classmethod
    def split_channels(cls, color: str) -> Tuple[MutableVector, float]:
        """Split channels."""

        start = 6
        channels = []
        alpha = 1.0
        for i, c in enumerate(parse.RE_CHAN_SPLIT.split(color[start:-1].strip()), 0):
            c = c.lower()
            if i <= 2:
                channels.append(cls.translate_channel(i, c))
            elif i == 3:
                alpha = cls.translate_channel(-1, c)
        return cls.null_adjust(channels, alpha)

    @classmethod
    def match(
        cls,
        string: str,
        start: int = 0,
        fullmatch: bool = True
    ) -> Optional[Tuple[Tuple[MutableVector, float], int]]:
        """Match a CSS color string."""

        match = super().match(string, start, fullmatch)
        if match is not None:
            return match
        m = cls.MATCH.match(string, start)
        if m is not None and (not fullmatch or m.end(0) == len(string)):
            return cls.split_channels(string[m.start(0):m.end(0)]), m.end(0)
        return None  # pragma: no cover
